# Author: Paolo Bonzini

import ijson.backends.python as ijson
import os, subprocess, sys

QEMU_IMG=os.environ.get('QEMU_IMG', '/usr/bin/qemu-img')
qemu_img_args = os.environ.get('QEMU_IMG', 'qemu-img').strip().split(' ')

class qemu_img:
    def __init__(self, *args):
        '''Run qemu-img and return the exit code'''
        devnull = open('/dev/null', 'r+')
        self.pipe = subprocess.Popen(qemu_img_args + list(args), stdin=devnull,
                            stdout=subprocess.PIPE,stderr=subprocess.PIPE,
                            close_fds=True)
        self.stdout = self.pipe.stdout

    def close(self):
        self.pipe.wait()
        (stdout, stderr) = self.pipe.communicate()
        if len(stderr):
            sys.stdout.write(stderr)
        if self.pipe.returncode > 0:
            sys.exit(self.pipe.returncode)

def copy_range(inf, inpos, nbytes, outf, outpos):
    inf.seek(inpos)
    outf.seek(outpos)
    while nbytes > 1048576:
        outf.write(inf.read(1048576))
        nbytes -= 1048576
    outf.write(inf.read(nbytes))

def progress(pos, size):
    sys.stdout.write('{0:.2f}%\r'.format(pos * 100.0 / size))
    sys.stdout.flush()

if len(sys.argv) == 1:
    print('Usage: python qemu-to-raw.py SOURCE DEST')
    print('')
    print('Converts SOURCE to raw format and writes it to DEST using qemu-img map')
    print('')
    sys.exit(1)

q = qemu_img('info', '--output=json', '--backing-chain', sys.argv[1])
files = []
n = 0
try:
    for prefix, event, value in ijson.parse(q.stdout):
        if event == 'end_map':
            n = n + 1
        elif prefix == 'item.virtual-size' and n == 0:
            size = value
        elif prefix == 'item.filename':
            f = open(value, 'r')
            files.append(f)
finally:
    q.close()

outf = open(sys.argv[2], 'w')
outf.truncate(size)

q = qemu_img('map', '--output=json', sys.argv[1])
names = []
try:
    for prefix, event, value in ijson.parse(q.stdout):
        if event == 'start_map':
            inpos = None
            outpos = None
            nbytes = None
            inf = None
            copy = True
        elif event == 'number' or event == 'boolean':
            if prefix == 'item.start':
                outpos = value
            elif prefix == 'item.length':
                nbytes = value
            elif prefix == 'item.offset':
                inpos = value
            elif prefix == 'item.depth':
                inf = files[value]
            elif prefix == 'item.zero':
                copy = copy and not value
            elif prefix == 'item.data':
                copy = copy and value
        elif event == 'end_map':
            if inpos is None:
                if copy:
                    raise Exception('Raw data not found in input file')
            else:
                progress(outpos,size)
                copy_range(inf, inpos, nbytes, outf, outpos)
            progress(outpos+nbytes,size)
finally:
    q.close()

sys.stdout.write('\n')
